import { match } from "@formatjs/intl-localematcher";
import { parse, TYPE } from "@formatjs/icu-messageformat-parser";
import { IntlMessageFormat } from "intl-messageformat";
function mergeImpl(a, b) {
  const result = { ...a };
  for (const key in b) {
    if (!(key in result) || typeof b[key] !== "object" || typeof result[key] !== "object") {
      result[key] = b[key];
    } else {
      result[key] = mergeImpl(result[key], b[key]);
    }
  }
  return result;
}
function mergeDicts(...parts) {
  return parts.reduce(mergeImpl, {});
}
function toDate(date, options = {}) {
  if (date instanceof Date) {
    return { date, options, type: "instant" };
  }
  if (!(date instanceof Object)) {
    return { date: date ? new Date(date) : /* @__PURE__ */ new Date(), options, type: "instant" };
  }
  if ("epochMilliseconds" in date) {
    return { date: new Date(date.epochMilliseconds), options, type: "instant" };
  } else {
    const isPlain = !date.timeZone;
    let iso = date.toString().replace(/\[.*\]/, "") + (isPlain ? "Z" : "");
    let type = "instant";
    if (isPlain) {
      options = { ...options, timeZone: "UTC" };
      type = "plainDateTime";
    }
    if (date.hour === void 0) {
      options = { ...options, timeStyle: void 0, hour: void 0, minute: void 0, second: void 0 };
      type = "plainDate";
    }
    if (date.month === void 0) {
      iso = `1970-01-01T${iso}`;
      options = { ...options, dateStyle: void 0, year: void 0, month: void 0, day: void 0 };
      type = "plainTime";
    }
    return {
      date: new Date(iso),
      options,
      type
    };
  }
}
function arrEquals(a, b) {
  if (!a || !b)
    return a === b;
  if (a.length !== b.length)
    return false;
  return a.every((value, index) => value === b[index]) ?? false;
}
function castArray(x = []) {
  if (x instanceof Array)
    return x;
  return [x];
}
function calcLocales(locale, fallbackToLessSpecific, fallback) {
  const requestedLocales = [locale];
  if (fallbackToLessSpecific) {
    let prefix = locale;
    while (prefix.includes("-")) {
      const index = prefix.lastIndexOf("-");
      prefix = prefix.slice(0, index);
      requestedLocales.push(prefix);
    }
  }
  if (fallback instanceof Function) {
    requestedLocales.push(...castArray(fallback(locale)));
  } else if (fallback) {
    requestedLocales.push(...castArray(fallback));
  }
  return requestedLocales;
}
function intlHelpers({
  cache,
  transform,
  dateTimeFormatOptions,
  listFormatOptions,
  numberFormatOptions,
  pluralRulesOptions,
  relativeTimeFormatOptions
}) {
  return {
    dateTimeFormat(date, options = dateTimeFormatOptions) {
      return transform((locale) => customDateTimeFormat(cache, locale, options, date));
    },
    dateTimeFormatRange(startDate, endDate, options = dateTimeFormatOptions) {
      return transform((locale) => customDateTimeFormatRange(cache, locale, options, startDate, endDate));
    },
    displayNames(code, options) {
      return transform((locale) => cache.get(Intl.DisplayNames, locale, options).of(code) ?? "");
    },
    listFormat(list, options = listFormatOptions) {
      return transform((locale) => cache.get(Intl.ListFormat, locale, options).format(list));
    },
    numberFormat(number, options = numberFormatOptions) {
      return transform((locale) => cache.get(Intl.NumberFormat, locale, options).format(number));
    },
    // numberFormatRange(start, end, options = numberFormatOptions) {
    //   return transform((locale) => cache.get(Intl.NumberFormat, locale, options).formatRange(start, end));
    // },
    pluralRules(number, options = pluralRulesOptions) {
      return transform((locale) => cache.get(Intl.PluralRules, locale, options).select(number));
    },
    relativeTimeFormat(value, unit, options = relativeTimeFormatOptions) {
      return transform((locale) => cache.get(Intl.RelativeTimeFormat, locale, options).format(value, unit));
    }
  };
}
function customDateTimeFormat(cache, locale, options_, date_) {
  const { date, options } = toDate(date_, options_) ?? {};
  return cache.get(Intl.DateTimeFormat, locale, options).format(date);
}
function customDateTimeFormatRange(cache, locale, options_, startDate_, endDate_) {
  const start = toDate(startDate_, options_);
  const end = toDate(endDate_);
  if (start.type !== end.type) {
    throw new Error("Intl.DateTimeFormat.formatRange accepts two values of the same type");
  }
  return cache.get(Intl.DateTimeFormat, locale, start.options).formatRange(start.date, end.date);
}
function resolveProvidedArgs(providedArgs) {
  return Object.fromEntries(
    Object.entries(providedArgs ?? {}).map(([id, value]) => {
      if (typeof value === "object" && value !== null && "get" in value) {
        return [id, value.get()];
      }
      return [id, value];
    })
  );
}
function hash(obj) {
  return JSON.stringify(obj, (_key, value) => {
    if (value instanceof Array || !(value instanceof Object))
      return value;
    const result = {};
    for (const key of Object.keys(value).sort()) {
      result[key] = value[key];
    }
    return result;
  });
}
class Cache {
  constructor(options) {
    this.options = options;
    this.cache = /* @__PURE__ */ new Map();
    this.delQueue = /* @__PURE__ */ new Map();
  }
  get(C, ...args) {
    var _a;
    if (((_a = this.options) == null ? void 0 : _a.maxEntries) === 0)
      return new C(...args);
    const key = hash(args);
    let innerCache = this.cache.get(C);
    if (!innerCache) {
      innerCache = /* @__PURE__ */ new Map();
      this.cache.set(C, innerCache);
    }
    let value = innerCache.get(key);
    if (!value) {
      this.clean();
      value = new C(...args);
      innerCache.set(key, value);
    }
    this.delQueue.delete(value);
    this.delQueue.set(value, [Date.now(), C, key]);
    return value;
  }
  clean() {
    const { maxEntries, ttl } = this.options ?? {};
    const now = Date.now();
    for (const [k, [t, k1, k2]] of this.delQueue.entries()) {
      if (maxEntries && maxEntries <= this.delQueue.size || ttl && t + ttl > now) {
        const innerCache = this.cache.get(k1);
        this.delQueue.delete(k);
        innerCache == null ? void 0 : innerCache.delete(k2);
        if ((innerCache == null ? void 0 : innerCache.size) === 0) {
          this.cache.delete(k1);
        }
      }
    }
  }
  clear() {
    this.cache.clear();
  }
}
function flattenDict(dict, path = "") {
  const flat = {};
  for (const [key, value] of Object.entries(dict)) {
    const newPath = path ? `${path}.${key}` : key;
    if (value === void 0)
      continue;
    if (value instanceof Object && !(value instanceof Array))
      Object.assign(flat, flattenDict(value, newPath));
    else
      flat[newPath] = value;
  }
  return flat;
}
class Store {
  constructor(options) {
    this.options = options;
    this.dicts = /* @__PURE__ */ new Map();
    this.cache = new Cache(this.options.cacheOptions);
    this.subs = /* @__PURE__ */ new Set();
  }
  load(locale) {
    let entry = this.dicts.get(locale);
    if (entry !== void 0)
      return entry;
    let dict = null;
    if (match([locale], [this.options.sourceLocale], "") && this.options.sourceDictionary) {
      dict = this.options.sourceDictionary;
    } else if (this.options.dicts instanceof Function) {
      dict = this.options.dicts(locale);
    } else if (this.options.dicts) {
      const availableLocales = Object.keys(this.options.dicts);
      const matching = match([locale], availableLocales, locale);
      dict = this.options.dicts[matching] ?? null;
      if (dict instanceof Function)
        dict = dict();
    }
    if (dict instanceof Promise) {
      entry = dict.then((resolvedDict) => {
        const flatDict = resolvedDict && flattenDict(resolvedDict);
        if (this.dicts.get(locale) === entry) {
          this.dicts.set(locale, flatDict);
        }
        return flatDict;
      }).catch(() => null);
    } else {
      entry = dict && flattenDict(dict);
    }
    this.dicts.set(locale, entry);
    if (entry instanceof Promise) {
      entry.then(() => this.notify());
    } else {
      this.notify();
    }
    return entry;
  }
  loadAll(...locales) {
    const dicts = locales.map((locale) => this.load(locale));
    if (dicts.some((dict) => dict instanceof Promise)) {
      return Promise.all(dicts).then((dicts2) => dicts2.filter(Boolean));
    }
    return dicts.filter(Boolean);
  }
  getAll(...locales) {
    const dicts = locales.map((locale) => this.load(locale));
    return dicts.filter(Boolean);
  }
  subscribe(locales, callback) {
    let last;
    const sub = () => {
      const dicts = this.getAll(...locales);
      if (!arrEquals(dicts, last)) {
        last = dicts;
        callback(dicts);
      }
    };
    sub();
    this.subs.add(callback);
    return () => {
      this.subs.delete(callback);
    };
  }
  clear() {
    this.dicts.clear();
    this.cache.clear();
    this.notify();
  }
  notify() {
    setTimeout(() => {
      for (const sub of this.subs)
        sub();
    });
  }
}
function mapPotentialArray(value, fn) {
  if (value instanceof Array)
    return value.map(fn);
  return fn(value);
}
function translate({
  dicts,
  sourceDict,
  id,
  values,
  fallback,
  fallbackIgnoresFallbackLocales,
  placeholder,
  locale,
  warn,
  cache,
  ignoreMissingArgs,
  providedArgs
}) {
  if (fallback !== void 0 && fallbackIgnoresFallbackLocales) {
    dicts = dicts.slice(0, 1);
  }
  const dict = dicts.find((dict2) => dict2 instanceof Promise || id in dict2);
  if (dict instanceof Promise) {
    return mapPotentialArray(
      sourceDict && !(sourceDict instanceof Promise) ? translate({
        dicts: [sourceDict],
        sourceDict,
        id,
        values,
        locale,
        cache,
        ignoreMissingArgs,
        providedArgs
      }) : void 0,
      (sourceTranslation) => {
        if (placeholder instanceof Function) {
          return placeholder(id, sourceTranslation);
        }
        return placeholder ?? "";
      }
    );
  }
  const template = dict == null ? void 0 : dict[id];
  if (!template) {
    if (fallback instanceof Function) {
      const sourceTranslation = sourceDict && !(sourceDict instanceof Promise) ? translate({
        dicts: [sourceDict],
        sourceDict,
        id,
        values,
        locale,
        cache,
        ignoreMissingArgs,
        providedArgs
      }) : void 0;
      return fallback(id, sourceTranslation);
    }
    if (fallback !== void 0)
      return fallback;
    warn == null ? void 0 : warn(locale, id);
    return id;
  }
  return mapPotentialArray(template, (template2) => format({ template: template2, values, locale, cache, ignoreMissingArgs, providedArgs }));
}
function format({
  template,
  values,
  locale,
  cache,
  ignoreMissingArgs,
  providedArgs
}) {
  values = { ...providedArgs, ...values };
  try {
    const ast = parse(template, { requiresOtherClause: false });
    const f = new IntlMessageFormat(ast, locale, void 0, {
      formatters: {
        getDateTimeFormat(locals, options) {
          const instance = cache.get(Intl.DateTimeFormat, locals, options);
          return {
            ...instance,
            format: (date) => customDateTimeFormat(cache, locale, options, date),
            formatRange: (startDate, endDate) => typeof startDate === "bigint" || typeof endDate === "bigint" ? instance.formatRange(startDate, endDate) : customDateTimeFormatRange(cache, locale, options, startDate, endDate)
          };
        },
        getNumberFormat(locals, options) {
          return cache.get(Intl.NumberFormat, locals, options);
        },
        getPluralRules(locals, options) {
          return cache.get(Intl.PluralRules, locals, options);
        }
      }
    });
    if (ignoreMissingArgs) {
      values = { ...values };
      for (const arg of new Set(findArgs(ast))) {
        if (!(arg in values)) {
          if (ignoreMissingArgs === true) {
            values[arg] = "";
          } else if (typeof ignoreMissingArgs === "string") {
            values[arg] = ignoreMissingArgs;
          } else {
            values[arg] = ignoreMissingArgs(arg, template);
          }
        }
      }
    }
    const msg = f.format(values);
    if (msg instanceof Array)
      return msg.join(" ");
    return String(msg);
  } catch (e) {
    return `Wrong format: ${String(e)}`;
  }
}
function findArgs(ast) {
  return ast.flatMap((el) => {
    if (el.type === TYPE.argument || el.type === TYPE.number || el.type === TYPE.date || el.type === TYPE.time) {
      return [el.value];
    }
    if (el.type === TYPE.select || el.type === TYPE.plural) {
      return [el.value, ...findArgs(Object.values(el.options).flatMap((option) => option.value))];
    }
    return [];
  });
}
const createGetTranslator = (store, {
  fallbackLocale,
  fallbackToLessSpecific = true,
  fallback: globalFallback,
  fallbackIgnoresFallbackLocales = false,
  warn,
  sourceLocale,
  dateTimeFormatOptions,
  listFormatOptions,
  numberFormatOptions,
  pluralRulesOptions,
  relativeTimeFormatOptions,
  ignoreMissingArgs,
  provideArgs
}) => async (locale) => {
  const dicts = await store.loadAll(locale, ...calcLocales(locale, fallbackToLessSpecific, fallbackLocale));
  const sourceDict = await store.load(sourceLocale);
  const t = (id, ...[values, options]) => {
    const fallback = (options == null ? void 0 : options.fallback) ?? globalFallback;
    return translate({
      dicts,
      sourceDict,
      id,
      values,
      fallback,
      fallbackIgnoresFallbackLocales,
      locale,
      warn,
      cache: store.cache,
      ignoreMissingArgs,
      providedArgs: resolveProvidedArgs(provideArgs)
    });
  };
  return Object.assign(t, {
    locale,
    unknown: t,
    dynamic: t,
    format(template, ...[values]) {
      return format({
        template,
        values,
        locale,
        cache: store.cache,
        ignoreMissingArgs,
        providedArgs: resolveProvidedArgs(provideArgs)
      });
    },
    ...intlHelpers({
      cache: store.cache,
      transform: (fn) => fn(locale),
      dateTimeFormatOptions,
      listFormatOptions,
      numberFormatOptions,
      pluralRulesOptions,
      relativeTimeFormatOptions
    })
  });
};
function createTranslator(options) {
  const store = new Store(options);
  return {
    getTranslator: createGetTranslator(store, options),
    clearDicts() {
      store.clear();
    }
  };
}
export {
  Store as S,
  arrEquals as a,
  calcLocales as b,
  createGetTranslator as c,
  castArray as d,
  createTranslator as e,
  format as f,
  hash as h,
  intlHelpers as i,
  mergeDicts as m,
  resolveProvidedArgs as r,
  translate as t
};
//# sourceMappingURL=index2.js.map
